import Cordova
import UIKit
import PointrKit

@objc(CDVPointrPlugin) class CDVPointrPlugin: CDVPlugin, UINavigationControllerDelegate, PTRDataManagerDelegate, PTRSiteManagerDelegate, PTRPositionManagerDelegate, PTRMapEventsListener {
    
    func log(_ message: String) {
        let dateFormatter = DateFormatter()
        dateFormatter.dateFormat = "[yyyy-MM-dd HH:mm:ss]"
        
        print("PointrPlugin -> \(dateFormatter.string(from: Date())) \(message)")
    }
    
    private var loggerLevel: PTRLoggerLevel = .info
    private var environment: PTREnvironment = .production
    private var licenseKey: String?
    private var dispatchGroupSiteManager = [DispatchGroup]()
    var mapDispatchGroupDataManager = [Int32: [DispatchGroup]]()
    private var dispatchGroupPositionManager = [DispatchGroup]()
    private var dispatchWorkItemAfterMapEndLoading: DispatchWorkItem?
    private var isRouteSummaryEnabled = true
    private var mapWidget: PTRMapWidgetViewController? = nil
    
    // region PTRMapEventsListener
    
    func mapDidEndLoading(_ map: PTRMapViewController) {
        map.removeListener(self)
        if let dispatchWorkItemAfterMapEndLoading = dispatchWorkItemAfterMapEndLoading {
            dispatchWorkItemAfterMapEndLoading.perform()
        }
        dispatchWorkItemAfterMapEndLoading = nil
    }
    
    // endregion
    
    // region PTRPositionManagerDelegate
    
    func onPositionManagerCalculatedLocation(_ calculatedLocation: PTRCalculatedLocation) {
        if (!calculatedLocation.isGeoValid()) {
            return
        }
        while(!dispatchGroupPositionManager.isEmpty) {
            let dispatchGroup = dispatchGroupPositionManager.removeLast()
            dispatchGroup.leave()
        }
    }

    // endregion
    
    // region PTRDataManagerDelegate
    
    func onDataManagerCompleteAll(for site: PTRSite, isSuccessful: Bool, dataFromOnline isOnlineData: Bool, errorMessages: [NSNumber : String]) {
        if (!isOnlineData) {
            return
        }
        while(self.mapDispatchGroupDataManager[site.internalIdentifier]?.isEmpty == false) {
            if let dispatchGroup = self.mapDispatchGroupDataManager[site.internalIdentifier]?.removeLast() {
                dispatchGroup.leave()
            }
        }
        
    }
    
    // endregioon PTRDataManagerDelegate
    
    
    // region PTRSiteManagerDelegate
    
    func onSiteManagerDataChanged() {
        while(!dispatchGroupSiteManager.isEmpty) {
            let dispatchGroup = dispatchGroupSiteManager.removeLast()
            dispatchGroup.leave()
        }
    }
    
    // endregion PTRSiteManagerDelegate
    
    public struct Location: Codable {
        var latitude: Double = -999999.0
        var longitude: Double = -999999.0
        var levelIndex: Int32 = -999999
        var buildingExternalIdentifier: String = ""
        var siteExternalIdentifier: String = ""
        var poiInternalIdentifier: String?
        var title: String?
        
        public func toJsonString() -> String? {
            let jsonEncoder = JSONEncoder()
            do {
                let data = try jsonEncoder.encode(self)
                return String(data: data, encoding: .utf8)
            }
            catch {}
            return nil
        }
        
        static func toLocation(_ ptrLocation: PTRLocation) -> Location {
            return Location(
                latitude: ptrLocation.coordinate.latitude,
                longitude: ptrLocation.coordinate.longitude,
                levelIndex: ptrLocation.level?.index ?? -999999,
                buildingExternalIdentifier: ptrLocation.building?.externalIdentifier ?? "",
                siteExternalIdentifier: ptrLocation.site?.externalIdentifier ?? ""
            )
        }
        
        static func toLocation(_ poi: PTRPoi) -> Location {
            var location = toLocation(poi.location)
            location.poiInternalIdentifier = poi.identifier
            location.title = poi.title
            return location
        }
        
        static func toLocation(_ string: String) -> Location? {
            let jsonDecoder = JSONDecoder()
            do {
                guard let data = string.data(using: .utf8) else {
                    return nil
                }
                let location = try jsonDecoder.decode(Location.self, from: data)
                return location
            } catch {}
            return nil
        }
    }
    
    // region actions
    
    @objc(initialize:)
    func initialize(command: CDVInvokedUrlCommand) {
        self.commandDelegate.run(inBackground: {
            guard command.arguments.count > 0,
                let licenseKey = command.arguments[0] as? String else {
                self.sendError(message: "Missing license key!!!", callbackId: command.callbackId)
                return
            }

            self.licenseKey = licenseKey
            var environment = "prod"
            if (command.arguments.count > 1) {
                environment = command.arguments[1] as? String ?? "prod"
            }
            switch(environment.lowercased()) {
            case "qa":
                self.environment = .QA
                break
            case "preprod":
                self.environment = .preProduction
                break
            case "dev":
                self.environment = .development
                break
            default:
                self.environment = .production
                break
            }
            if (command.arguments.count > 2) {
                self.loggerLevel = PTRLoggerLevel(rawValue: command.arguments[2] as? Int32 ?? 4) ?? .info
            }
            self.commandDelegate.send(CDVPluginResult(status: CDVCommandStatus.ok, messageAs: "SUCCESS"), callbackId: command.callbackId)
        })
    }
    
    @objc(start:)
    func start(command: CDVInvokedUrlCommand) {
        let state = self.start()
        if(state != .running) {
            self.sendError(message: "\(PTRPointrStateToString(Pointr.shared.state))", callbackId: command.callbackId)
            return
        }
        self.commandDelegate.send(CDVPluginResult(status: CDVCommandStatus.ok, messageAs: "SUCCESS"), callbackId: command.callbackId)
    }
    
    @objc(showPoiDetails:)
    func showPoiDetails(command: CDVInvokedUrlCommand) {
        let state = self.start()
        if(state != .running) {
            self.sendError(message: "\(PTRPointrStateToString(Pointr.shared.state))", callbackId: command.callbackId)
            return
        }
        
        DispatchQueue.global().async {
            
            guard let dataManager = Pointr.shared.dataManager,
                  let siteManager = Pointr.shared.siteManager,
                  let poiManager = Pointr.shared.poiManager else {
                self.sendError(message: "Managers are not ready", callbackId: command.callbackId)
                return
            }
            
            guard command.arguments.count > 1,
                  let siteExternalIdentifier = command.arguments[0] as? String,
                  let poiInternalIdentifier: String = command.arguments[1] as? String else {
                self.sendError(message: "Missing argument: siteExternalIdentifier and poiInternalIdentifier", callbackId: command.callbackId)
                return
            }
            
            guard let site = self.getSite(with: siteExternalIdentifier, siteManager: siteManager) else {
                self.sendError(message: "Could not find site \(siteExternalIdentifier)", callbackId: command.callbackId)
                return
            }
            
            if !self.loadData(for: site, dataManager: dataManager) {
                self.log("Data are not ready for site: \(siteExternalIdentifier)")
            }
            
            guard let poi = poiManager.pois(for: site, withIdentifier: poiInternalIdentifier) else {
                self.sendError(message: "Could not find poi \(poiInternalIdentifier)", callbackId: command.callbackId)
                return
            }
            
            DispatchQueue.main.async {
                let mapWidget = self.configureMapWidget()
                self.presentMapWidget(mapWidget)
                mapWidget.loadAndSet(site: site)
                mapWidget.showPoiDetails(poi)
                self.commandDelegate.send(CDVPluginResult(status: CDVCommandStatus.ok, messageAs: "SUCCESS"), callbackId: command.callbackId)
            }
            
        }
    }
    
    @objc(highlightPoi:)
    func highlightPoi(command: CDVInvokedUrlCommand) {
        let state = self.start()
        if(state != .running) {
            self.sendError(message: "\(PTRPointrStateToString(Pointr.shared.state))", callbackId: command.callbackId)
            return
        }
        
        guard command.arguments.count > 1,
              let siteExternalIdentifier = command.arguments[0] as? String,
              let poiInternalIdentifier: String = command.arguments[1] as? String else {
            self.sendError(message: "Missing argument: siteExternalIdentifier and poiInternalIdentifier", callbackId: command.callbackId)
            return
        }
        
        DispatchQueue.global().async {
            
            guard let dataManager = Pointr.shared.dataManager,
                  let siteManager = Pointr.shared.siteManager,
                  let poiManager = Pointr.shared.poiManager else {
                self.sendError(message: "Managers are not ready", callbackId: command.callbackId)
                return
            }
            
            guard let site = self.getSite(with: siteExternalIdentifier, siteManager: siteManager) else {
                self.sendError(message: "Could not find site \(siteExternalIdentifier)", callbackId: command.callbackId)
                return
            }
            
            if !self.loadData(for: site, dataManager: dataManager) {
                self.log("Data are not ready for site: \(siteExternalIdentifier)")
            }
            
            guard let poi = poiManager.pois(for: site, withIdentifier: poiInternalIdentifier) else {
                self.sendError(message: "Could not find poi \(poiInternalIdentifier)", callbackId: command.callbackId)
                return
            }
            
            DispatchQueue.main.async {
                let mapWidget = self.configureMapWidget()
                self.presentMapWidget(mapWidget)
                if let level = poi.location.level {
                    mapWidget.loadAndSet(level: level)
                    self.dispatchWorkItemAfterMapEndLoading = DispatchWorkItem {
                        self.highlightPoi(map: mapWidget.mapViewController, poi: poi, callbackId: command.callbackId)
                    }
                    if mapWidget.mapViewController.isMapFullyRendered {
                        self.dispatchWorkItemAfterMapEndLoading?.perform()
                        self.dispatchWorkItemAfterMapEndLoading = nil
                    } else {
                        mapWidget.mapViewController.addListener(self)
                    }
                } else {
                    self.sendError(message: "Invalid level for poi: \(poiInternalIdentifier)", callbackId: command.callbackId)
                }
            }
            
        }
    }
    
    @objc(showPathfindingToPoi:)
    func showPathfindingToPoi(command: CDVInvokedUrlCommand) {
        
        let state = self.start()
        if(state != .running) {
            self.sendError(message: "\(PTRPointrStateToString(Pointr.shared.state))", callbackId: command.callbackId)
            return
        }
        DispatchQueue.global().async {
            
            guard let dataManager = Pointr.shared.dataManager,
                  let siteManager = Pointr.shared.siteManager,
                  let poiManager = Pointr.shared.poiManager,
                  let positionManager = Pointr.shared.positionManager,
                  let pathManager = Pointr.shared.pathManager else {
                self.sendError(message: "Managers are not ready", callbackId: command.callbackId)
                return
            }
            
            guard command.arguments.count > 1,
                  let siteExternalIdentifier = command.arguments[0] as? String,
                  let poiInternalIdentifier: String = command.arguments[1] as? String else {
                self.sendError(message: "Missing argument: siteExternalIdentifier and poiInternalIdentifier", callbackId: command.callbackId)
                return
            }
            
            guard let site = self.getSite(with: siteExternalIdentifier, siteManager: siteManager) else {
                self.sendError(message: "Could not find site \(siteExternalIdentifier)", callbackId: command.callbackId)
                return
            }
            
            if !self.loadData(for: site, dataManager: dataManager) {
                self.log("Data are not ready for site: \(siteExternalIdentifier)")
            }
            
            guard let poi = poiManager.pois(for: site, withIdentifier: poiInternalIdentifier) else {
                self.sendError(message: "Could not find poi \(poiInternalIdentifier)", callbackId: command.callbackId)
                return
            }
            guard let location = self.getCurrentLocation(positionManager) else {
                self.sendError(message: "Could not find your location", callbackId: command.callbackId)
                return
            }
            guard location.isValid(), let currentLevel = location.level else {
                self.sendError(message: "Your location is not within a site", callbackId: command.callbackId)
                return
            }
            guard let pathSession = pathManager.startPathSession(withDestinations: [poi.location]) else {
                self.sendError(message: "Could not create path from \(location) to \(poi)", callbackId: command.callbackId)
                return
            }
            
            DispatchQueue.main.async {
                let mapWidget = self.configureMapWidget()
                self.presentMapWidget(mapWidget)
                mapWidget.loadAndSet(level: currentLevel)
                self.dispatchWorkItemAfterMapEndLoading = DispatchWorkItem {
                    self.zoomToCoordinate(map: mapWidget.mapViewController, coordinate: location.coordinate)
                    mapWidget.showPathFinding(pathSession: pathSession, destination: poi)
                    self.commandDelegate.send(CDVPluginResult(status: CDVCommandStatus.ok, messageAs: "SUCCESS"), callbackId: command.callbackId)
                }
                if mapWidget.mapViewController.isMapFullyRendered {
                    self.dispatchWorkItemAfterMapEndLoading?.perform()
                    self.dispatchWorkItemAfterMapEndLoading = nil
                } else {
                    mapWidget.mapViewController.addListener(self)
                    
                }
            }
            
        }
    }
    
    @objc(showPathfindingToLocation:)
    func showPathfindingToLocation(command: CDVInvokedUrlCommand) {
        
        let state = self.start()
        if(state != .running) {
            self.sendError(message: "\(PTRPointrStateToString(Pointr.shared.state))", callbackId: command.callbackId)
            return
        }
        
        
        DispatchQueue.global().async {
            
            guard command.arguments.count > 0 else {
                self.sendError(message: "No command arguments found", callbackId: command.callbackId)
                return
            }
            guard let locationString = command.arguments[0] as? String else {
                self.sendError(message: "Invalid argument \(command.arguments[0])", callbackId: command.callbackId)
                return
            }
            
            guard let location = Location.toLocation(locationString) else {
                self.sendError(message: "Not a valid location json string: \(locationString)", callbackId: command.callbackId)
                return
            }
            
            guard location.levelIndex != -999999 else {
                self.sendError(message: "Destination is not within a site. Cannot navigate", callbackId: command.callbackId)
                return
            }
            
            
            guard let dataManager = Pointr.shared.dataManager,
                  let poiManager = Pointr.shared.poiManager,
                  let positionManager = Pointr.shared.positionManager,
                  let pathManager = Pointr.shared.pathManager else {
                self.sendError(message: "Managers are not ready", callbackId: command.callbackId)
                return
            }
            
            guard let currentLocation = self.getCurrentLocation(positionManager) else {
                self.sendError(message: "Could not find your location", callbackId: command.callbackId)
                return
            }
            guard currentLocation.isValid(),
                  let site = currentLocation.site else {
                self.sendError(message: "You are not within a site", callbackId: command.callbackId)
                return
            }
            guard site.externalIdentifier == location.siteExternalIdentifier else {
                self.sendError(message: "You are not on the same site with the given location: \(site)", callbackId: command.callbackId)
                return
            }
            let coordinate = CLLocationCoordinate2D(latitude: location.latitude, longitude: location.longitude)
            guard let building = site.buildings.filter({ $0.externalIdentifier == location.buildingExternalIdentifier }).first else {
                self.sendError(message: "Unable to find building: \(location.buildingExternalIdentifier) site: \(location.siteExternalIdentifier)", callbackId: command.callbackId)
                return
            }
            guard let level = building.levels.filter({ $0.index == location.levelIndex }).first else {
                self.sendError(message: "Unable to find level: \(location.levelIndex) building: \(location.buildingExternalIdentifier) site: \(location.siteExternalIdentifier)", callbackId: command.callbackId)
                return
            }
            guard let ptrLocation = PTRLocation(coordinate: coordinate, level: level) else {
                self.sendError(message: "Could not create location to \(location)", callbackId: command.callbackId)
                return
            }
            
            guard self.loadData(for: site, dataManager: dataManager) else {
                self.sendError(message: "Data is not ready for site: \(site.externalIdentifier)", callbackId: command.callbackId)
                return
            }
            
            var poi: PTRPoi?
            if let poiInternalIdentifier = location.poiInternalIdentifier {
                poi = poiManager.pois(for: site, withIdentifier: poiInternalIdentifier)
            } else {
                poi = poiManager.pois(for: level)?.getPoiList().sorted(by: {
                    pathManager.approximateDistance(from: ptrLocation, to: $0.location) < pathManager.approximateDistance(from: ptrLocation, to: $1.location)
                }).first
            }
            
            guard let destination = poi else {
                self.sendError(message: "Could not find a poi close to the location at level \(String (describing: ptrLocation.level))", callbackId: command.callbackId)
                return
            }
            
            guard let pathSession = pathManager.startPathSession(withDestinations: [ptrLocation]) else {
                self.sendError(message: "Could not create path to location: \(ptrLocation)", callbackId: command.callbackId)
                return
            }
            guard let currentLevel = currentLocation.level else {
                self.sendError(message: "You are not in a valid level", callbackId: command.callbackId)
                return
            }
            
            DispatchQueue.main.async {
                let mapWidget = self.configureMapWidget()
                self.presentMapWidget(mapWidget)
                mapWidget.loadAndSet(level: currentLevel)
                self.dispatchWorkItemAfterMapEndLoading = DispatchWorkItem {
                    self.zoomToCoordinate(map: mapWidget.mapViewController, coordinate: currentLocation.coordinate)
                    mapWidget.showPathFinding(pathSession: pathSession, destination: destination)
                    self.commandDelegate.send(CDVPluginResult(status: CDVCommandStatus.ok, messageAs: "SUCCESS"), callbackId: command.callbackId)
                }
                if mapWidget.mapViewController.isMapFullyRendered {
                    self.dispatchWorkItemAfterMapEndLoading?.perform()
                    self.dispatchWorkItemAfterMapEndLoading = nil
                } else {
                    mapWidget.mapViewController.addListener(self)
                }
            }
        }
    }
    
    @objc(showPathfindingBetweenPois:)
    func showPathfindingBetweenPois(command: CDVInvokedUrlCommand) {
        let state = self.start()
        if(state != .running) {
            self.sendError(message: "\(PTRPointrStateToString(Pointr.shared.state))", callbackId: command.callbackId)
            return
        }
        DispatchQueue.global().async {
            
            
            guard let dataManager = Pointr.shared.dataManager,
                  let siteManager = Pointr.shared.siteManager,
                  let poiManager = Pointr.shared.poiManager,
                  let pathManager = Pointr.shared.pathManager else {
                self.sendError(message: "Managers are not ready", callbackId: command.callbackId)
                return
            }
            
            guard command.arguments.count > 2,
                  let siteExternalIdentifier = command.arguments[0] as? String,
                  let poiInternalIdentifier1 = command.arguments[1] as? String,
                  let poiInternalIdentifier2 = command.arguments[2] as? String else {
                self.sendError(message: "Missing argument: siteExternalIdentifier, poiInternalIdentifier1, poiInternalIdentifier2", callbackId: command.callbackId)
                return
            }
            
            guard let site = self.getSite(with: siteExternalIdentifier, siteManager: siteManager) else {
                self.sendError(message: "Could not find site \(siteExternalIdentifier)", callbackId: command.callbackId)
                return
            }
            
            if !self.loadData(for: site, dataManager: dataManager) {
                self.log("Data are not ready for site: \(siteExternalIdentifier)")
            }
            
            guard let from = poiManager.pois(for: site, withIdentifier: poiInternalIdentifier1) else {
                self.sendError(message: "Could not find poi: \(poiInternalIdentifier1) within site: \(siteExternalIdentifier)", callbackId: command.callbackId)
                return
            }
            
            guard let to = poiManager.pois(for: site, withIdentifier: poiInternalIdentifier2) else {
                self.sendError(message: "Could not find poi: \(poiInternalIdentifier2) within site: \(siteExternalIdentifier)", callbackId: command.callbackId)
                return
            }
            
            guard let path = pathManager.calculatePath(fromLocation: from.location, toNearestLocationIn: [to.location]) else {
                self.sendError(message: "Unable to calculate path between locations", callbackId: command.callbackId)
                return
            }
            
            guard let fromLevel = from.location.level else {
                self.sendError(message: "Starting poi is not within any level", callbackId: command.callbackId)
                return
            }
            
            DispatchQueue.main.async {
                let mapWidget = self.configureMapWidget()
                self.presentMapWidget(mapWidget)
                mapWidget.loadAndSet(level: fromLevel)
                self.dispatchWorkItemAfterMapEndLoading = DispatchWorkItem {
                    self.zoomToCoordinate(map: mapWidget.mapViewController, coordinate: from.coordinate)
                    mapWidget.mapViewController.currentPath = path
                    // to use different images for source and destination pois, you have to add new layers for each image
//                    let sourceLayer = PTRMapViewAnnotationLayer(identifier: UUID().uuidString, icon: UIImage.ptr_imageNamed("pf-direction-follow-the-line"), shouldAllowOverlap: true)
//                    let destinationLayer = PTRMapViewAnnotationLayer(identifier: UUID().uuidString, icon: UIImage.ptr_imageNamed("pf-direction-destination-flag"), shouldAllowOverlap: true)
                    let sourceAnnotation = PTRMapViewAnnotation(coordinate: from.coordinate, mapLevel: from.mapLevel, text: from.title)
                    let destinationAnnotation = PTRMapViewAnnotation(coordinate: to.coordinate, mapLevel: to.mapLevel, text: to.title)
                    mapWidget.mapViewController.addAnnotation(sourceAnnotation)//, toLayer: sourceLayer)
                    mapWidget.mapViewController.addAnnotation(destinationAnnotation)//, toLayer: destinationLayer)
                    self.commandDelegate.send(CDVPluginResult(status: CDVCommandStatus.ok, messageAs: "SUCCESS"), callbackId: command.callbackId)
                }
                if mapWidget.mapViewController.isMapFullyRendered {
                    self.dispatchWorkItemAfterMapEndLoading?.perform()
                    self.dispatchWorkItemAfterMapEndLoading = nil
                } else {
                    mapWidget.mapViewController.addListener(self)
                }
                
            }
            
        }
    }
    
    @objc(clearPath:)
    func clearPath(command: CDVInvokedUrlCommand) {
        if Pointr.shared.state == .running {
            let mapWidget = self.configureMapWidget()
            mapWidget.mapViewController.currentPath = nil
            mapWidget.mapViewController.removeAllAnnotations()
        }
        self.commandDelegate.send(CDVPluginResult(status: CDVCommandStatus.ok, messageAs: "SUCCESS"), callbackId: command.callbackId)
    }
    
    @objc(setPathfindingMode:)
    func setPathfindingMode(command: CDVInvokedUrlCommand) {
        let state = self.start()
        if(state != .running) {
            self.sendError(message: "\(PTRPointrStateToString(Pointr.shared.state))", callbackId: command.callbackId)
            return
        }
        
        DispatchQueue.global().async {
            
            guard let pathManager = Pointr.shared.pathManager else {
                self.sendError(message: "Managers are not ready", callbackId: command.callbackId)
                return
            }
            
            guard command.arguments.count > 0,
                  let mode = command.arguments[0] as? Int32 else {
                self.sendError(message: "Missing argument: Path finding mode", callbackId: command.callbackId)
                return
            }
            
            pathManager.setCurrentMode(PTRPathfindingMode(rawValue: mode) ?? PTRPathfindingMode.normal)
            
            self.commandDelegate.send(CDVPluginResult(status: CDVCommandStatus.ok, messageAs: "SUCCESS"), callbackId: command.callbackId)
        }
    }
    
    @objc(getCurrentLocation:)
    func getCurrentLocation(command: CDVInvokedUrlCommand) {
        let state = self.start()
        if state != .running {
            self.sendError(message: "\(PTRPointrStateToString(state))", callbackId: command.callbackId)
            return
        }
        DispatchQueue.global().async {
            var poiTypeName: String? = nil
            if (command.arguments.count > 0) {
                poiTypeName = command.arguments[0] as? String
            }
            var timeoutInSeconds: Int = 60
            if (command.arguments.count > 1) {
                timeoutInSeconds = Int(command.arguments[1] as? Int32 ?? 60)
            }
            
            guard let positionManager = Pointr.shared.positionManager,
                  let poiManager = Pointr.shared.poiManager else {
                self.sendError(message: "Managers are not ready", callbackId: command.callbackId)
                return
            }
            guard let ptrLocation = self.getCurrentLocation(positionManager, timeoutInSeconds: timeoutInSeconds) else {
                self.sendError(message: "Could not find your location", callbackId: command.callbackId)
                return
            }
            var location = Location.toLocation(ptrLocation)
            guard let poiTypeName = poiTypeName, !poiTypeName.isEmpty else {
                self.commandDelegate.send(CDVPluginResult(status: CDVCommandStatus.ok, messageAs: location.toJsonString()), callbackId: command.callbackId)
                return
            }
            if let level = ptrLocation.level,
               let poi = poiManager.pois(for: level, withTypes: [poiTypeName], shouldSortByProximity: true)?.getPoiList().first  {
                location.poiInternalIdentifier = poi.identifier
                location.title = poi.title
            } else {
                self.log("Could not find a poi close to your location: \(String (describing: ptrLocation.level))")
            }
            self.commandDelegate.send(CDVPluginResult(status: CDVCommandStatus.ok, messageAs: location.toJsonString()), callbackId: command.callbackId)
        }
    }
    
    @objc(showMap:)
    func showMap(command: CDVInvokedUrlCommand) {
        
        let state = self.start()
        if(state != .running) {
            self.sendError(message: "\(PTRPointrStateToString(Pointr.shared.state))", callbackId: command.callbackId)
            return
        }
        DispatchQueue.global().async {
            
            guard let siteManager = Pointr.shared.siteManager else {
                self.sendError(message: "Managers are not ready", callbackId: command.callbackId)
                return
            }
            
            let sites = siteManager.sites()
            guard !sites.isEmpty else {
                self.sendError(message: "No sites available", callbackId: command.callbackId)
                return
            }
            
            guard let site = sites.first else {
                self.sendError(message: "No sites available", callbackId: command.callbackId)
                return
            }
            
            DispatchQueue.main.async {
                let mapWidget = self.configureMapWidget()
                self.presentMapWidget(mapWidget)
                mapWidget.loadAndSet(site: site)
                self.commandDelegate.send(CDVPluginResult(status: CDVCommandStatus.ok, messageAs: "SUCCESS"), callbackId: command.callbackId)
            }
            
        }
    }

    @objc(showSite:)
    func showSite(command: CDVInvokedUrlCommand) {
        
        let state = self.start()
        if(state != .running) {
            self.sendError(message: "\(PTRPointrStateToString(Pointr.shared.state))", callbackId: command.callbackId)
            return
        }
        DispatchQueue.global().async {
            
            guard let siteManager = Pointr.shared.siteManager else {
                self.sendError(message: "Managers are not ready", callbackId: command.callbackId)
                return
            }
            
            guard command.arguments.count > 0,
                  let siteExternalIdentifier = command.arguments[0] as? String else {
                self.sendError(message: "Missing argument: siteExternalIdentifier", callbackId: command.callbackId)
                return
            }
            
            guard let site = self.getSite(with: siteExternalIdentifier, siteManager: siteManager) else {
                self.sendError(message: "Could not find site \(siteExternalIdentifier)", callbackId: command.callbackId)
                return
            }
            
            DispatchQueue.main.async {
                let mapWidget = self.configureMapWidget()
                self.presentMapWidget(mapWidget)
                mapWidget.loadAndSet(site: site)
                self.commandDelegate.send(CDVPluginResult(status: CDVCommandStatus.ok, messageAs: "SUCCESS"), callbackId: command.callbackId)
            }
            
        }
    }
    
    @objc(showBuilding:)
    func showBuilding(command: CDVInvokedUrlCommand) {
        
        let state = self.start()
        if(state != .running) {
            self.sendError(message: "\(PTRPointrStateToString(Pointr.shared.state))", callbackId: command.callbackId)
            return
        }
        DispatchQueue.global().async {
            
            guard let siteManager = Pointr.shared.siteManager else {
                self.sendError(message: "Managers are not ready", callbackId: command.callbackId)
                return
            }
            
            guard command.arguments.count > 1,
                  let siteExternalIdentifier = command.arguments[0] as? String,
                  let buildingExternalIdentifier = command.arguments[1] as? String else {
                self.sendError(message: "Missing argument: siteExternalIdentifier and buildingExternalIdentifier", callbackId: command.callbackId)
                return
            }
            
            
            guard let site = self.getSite(with: siteExternalIdentifier, siteManager: siteManager) else {
                self.sendError(message: "Could not find site \(siteExternalIdentifier)", callbackId: command.callbackId)
                return
            }
            guard let building = site.buildings.filter({ $0.externalIdentifier == buildingExternalIdentifier }).first else {
                self.sendError(message: "Could not find building \(buildingExternalIdentifier)", callbackId: command.callbackId)
                return
            }
            DispatchQueue.main.async {
                let mapWidget = self.configureMapWidget()
                self.presentMapWidget(mapWidget)
                mapWidget.loadAndSet(building: building)
                self.commandDelegate.send(CDVPluginResult(status: CDVCommandStatus.ok, messageAs: "SUCCESS"), callbackId: command.callbackId)
            }
            
        }
    }
    
    @objc(showLevel:)
    func showLevel(command: CDVInvokedUrlCommand) {
        
        let state = self.start()
        if(state != .running) {
            self.sendError(message: "\(PTRPointrStateToString(Pointr.shared.state))", callbackId: command.callbackId)
            return
        }
        DispatchQueue.global().async {
            
            guard let siteManager = Pointr.shared.siteManager else {
                self.sendError(message: "Managers are not ready", callbackId: command.callbackId)
                return
            }
            
            guard command.arguments.count > 2,
                  let siteExternalIdentifier = command.arguments[0] as? String,
                  let buildingExternalIdentifier = command.arguments[1] as? String,
                  let levelIndex = command.arguments[2] as? Int32 else {
                self.sendError(message: "Missing argument: siteExternalIdentifier, buildingExternalIdentifier and levelIndex", callbackId: command.callbackId)
                return
            }
            guard let site = self.getSite(with: siteExternalIdentifier, siteManager: siteManager) else {
                self.sendError(message: "Could not find site \(siteExternalIdentifier)", callbackId: command.callbackId)
                return
            }
            guard let building = site.buildings.filter({ $0.externalIdentifier == buildingExternalIdentifier }).first else {
                self.sendError(message: "Could not find building \(buildingExternalIdentifier)", callbackId: command.callbackId)
                return
            }
            guard let level = building.levels.filter({ $0.index == levelIndex }).first else {
                self.sendError(message: "Could not find level: \(levelIndex)", callbackId: command.callbackId)
                return
            }
            
            DispatchQueue.main.async {
                let mapWidget = self.configureMapWidget()
                self.presentMapWidget(mapWidget)
                mapWidget.loadAndSet(level: level)
                self.commandDelegate.send(CDVPluginResult(status: CDVCommandStatus.ok, messageAs: "SUCCESS"), callbackId: command.callbackId)
            }
            
        }
    }
    
    @objc(loadDataForSite:)
    func loadDataForSite(command: CDVInvokedUrlCommand) {
        
        let state = self.start()
        if(state != .running) {
            self.sendError(message: "\(PTRPointrStateToString(Pointr.shared.state))", callbackId: command.callbackId)
            return
        }
        DispatchQueue.global().async {
            
            guard let dataManager = Pointr.shared.dataManager,
                  let siteManager = Pointr.shared.siteManager else {
                self.sendError(message: "Managers are not ready", callbackId: command.callbackId)
                return
            }
            
            guard command.arguments.count > 0,
                  let siteExternalIdentifier = command.arguments[0] as? String else {
                self.sendError(message: "Missing argument: siteExternalIdentifier", callbackId: command.callbackId)
                return
            }
            
            guard let site = self.getSite(with: siteExternalIdentifier, siteManager: siteManager) else {
                self.sendError(message: "Could not find site \(siteExternalIdentifier)", callbackId: command.callbackId)
                return
            }
            
            if self.loadData(for: site, dataManager: dataManager) {
                self.commandDelegate.send(CDVPluginResult(status: CDVCommandStatus.ok, messageAs: "SUCCESS"), callbackId: command.callbackId)
                return
            }
            self.sendError(message: "Data are not ready for site: \(siteExternalIdentifier)", callbackId: command.callbackId)
        }
    }
    
    @objc(stop:)
    func stop(command: CDVInvokedUrlCommand) {
        if(Pointr.shared.state == .running) {
            Pointr.shared.positionManager?.removeListener(self)
            Pointr.shared.dataManager?.removeListener(self)
            Pointr.shared.siteManager?.removeListener(self)
        }
        Pointr.shared.stop()
    }
    
    @objc(shouldHideRouteSummaryView:)
    func shouldHideRouteSummaryView(command: CDVInvokedUrlCommand) {
        if !command.arguments.isEmpty {
            isRouteSummaryEnabled = !(command.arguments[0] as? Bool ?? false)
        }
        _ = configureMapWidget()
        self.commandDelegate.send(CDVPluginResult(status: CDVCommandStatus.ok, messageAs: "SUCCESS"), callbackId: command.callbackId)
    }
    
    // end of actions
    
    func start() -> PointrState {
        guard let licenseKey = licenseKey else {
            return .off
        }
        let params = PTRParams()
        params.mode = PointrDebugMode()
        params.licenseKey = licenseKey
        params.environment = self.environment
        params.loggerLevel = self.loggerLevel
        if(Pointr.shared.state == .running) {
            return .running
        }
        let dispatchGroup = DispatchGroup()
        dispatchGroup.enter()
        Pointr.shared.start(with: params) { state in
            self.log("\(PTRPointrStateToString(state))")
            if (state == .running || state.rawValue <= 0) {
                dispatchGroup.leave()
            }
        }
        dispatchGroup.wait()
        if (Pointr.shared.state == .running) {
            self.log("Mobile SDK version \(Pointr.shared.version)")
            Pointr.shared.positionManager?.addListener(self)
        }
        return Pointr.shared.state
    }
    
    func getCurrentLocation(_ positionManager: PTRPositionManagerInterface, timeoutInSeconds: Int = 15) -> PTRCalculatedLocation? {
        if let ptrLocation = positionManager.currentCalculatedLocation {
            if ptrLocation.isValid() {
                return ptrLocation
            }
        }
        if (timeoutInSeconds <= 0) {
            return nil
        }
        let dispatchGroup = DispatchGroup()
        self.dispatchGroupPositionManager.append(dispatchGroup)
        dispatchGroup.enter()
        let result = dispatchGroup.wait(wallTimeout: .now() + DispatchTimeInterval.seconds(timeoutInSeconds))
        if result == .success {
            return positionManager.currentCalculatedLocation
        }
        return nil
    }
    
    func getSite(with siteExternalIdentifier: String, siteManager: PTRSiteManagerInterface, timeoutInSeconds: Int = 15) -> PTRSite? {
        if let site = siteManager.site(withExternalIdentifier: siteExternalIdentifier) {
            return site
        } else {
            let dispatchGroup = DispatchGroup()
            self.dispatchGroupSiteManager.append(dispatchGroup)
            dispatchGroup.enter()
            siteManager.addListener(self)
            _ = dispatchGroup.wait(wallTimeout: .now() + DispatchTimeInterval.seconds(timeoutInSeconds))
            siteManager.removeListener(self)
        }
        return siteManager.site(withExternalIdentifier: siteExternalIdentifier)
    }
    
    func loadData(for site: PTRSite, dataManager: PTRDataManagerInterface) -> Bool {
        if(dataManager.isContentReady(forSite: site.internalIdentifier)) {
            return true
        }
        let dispatchGroup = DispatchGroup()
        self.mapDispatchGroupDataManager[site.internalIdentifier] = self.mapDispatchGroupDataManager[site.internalIdentifier] ?? [DispatchGroup]()
        self.mapDispatchGroupDataManager[site.internalIdentifier]?.append(dispatchGroup)
        dataManager.addListener(self)
        dispatchGroup.enter()
        dataManager.loadData(forSite: site.internalIdentifier)
        dispatchGroup.wait()
        dataManager.removeListener(self)
        return dataManager.isContentReady(forSite: site.internalIdentifier)
    }
    
    func sendError(message: String, callbackId: String) {
        self.log("\(message)")
        commandDelegate.send(CDVPluginResult(status: CDVCommandStatus.error, messageAs: message), callbackId: callbackId)
    }
    
    func configureMapWidget () -> PTRMapWidgetViewController {
        let pcConfiguration = PTRMutableConfiguration()
        pcConfiguration.userInterfaceConfiguration.shouldFocusOnFirstUserLocation = false
        Pointr.shared.configurationManager?.updateRuntimeConfiguration(pcConfiguration)

        let configuration = PTRMapWidgetConfiguration()
        configuration.isJoystickEnabled = false
        configuration.isAugmentingRealityEnabled = false
        configuration.isRouteSummaryEnabled = self.isRouteSummaryEnabled
        mapWidget = PTRMapWidgetViewController(configuration: configuration)
        return mapWidget!
    }
    
    func presentMapWidget(_ mapWidget: PTRMapWidgetViewController) {
        mapWidget.modalPresentationStyle = .fullScreen
        self.viewController.present(mapWidget, animated: true)
    }
    
    func zoomToCoordinate(map: PTRMapViewController, coordinate: CLLocationCoordinate2D) {
        map.zoomToCoordinate(coordinate)
    }
    
    func highlightPoi(map: PTRMapViewController, poi: PTRPoi, callbackId: String) {
        map.highlightPoi(poi)
        map.zoomToCoordinate(poi.coordinate)
        self.commandDelegate.send(CDVPluginResult(status: CDVCommandStatus.ok, messageAs: "SUCCESS"), callbackId: callbackId)
    }
}
